;;; more-vc.el
;;;
;;; Copyright (C) 2007, 2008 Thien-Thi Nguyen
;;;
;;; This file is part of ttn's personal elisp library, released under
;;; the terms of the GNU General Public License as published by the
;;; Free Software Foundation; either version 3, or (at your option) any
;;; later version.  There is NO WARRANTY.  See file COPYING for details.
;;;
;;; Description: Additional version-control functionality.

;;; Plug holes.

(require 'scheme)
(eval-when-compile (require 'cl))
(require 'vc)
(eval-when-compile (require 'vc-git))
(eval-when-compile (require 'log-view))
(eval-when-compile (require 'diff-mode))

;;;###autoload
(defun ttn-vc-toggle-read-only ()
  (interactive)
  (if (or (and (boundp 'vc-dired-mode) vc-dired-mode)
          ;; use boundp because vc.el might not be loaded
          (vc-backend buffer-file-name))
      (vc-next-action nil)
    (toggle-read-only)))


;;; New functionality.

;;;###autoload
(defun more-vc-set-state (files state)
  "Set \"file state\" to STATE for each file in FILES.
Revert any buffers visiting those files and return them in a list.
The backend to use is determined by the first element in FILES."
  (let* ((backend (or (vc-backend (expand-file-name (car files)))
                      (error "Not under version control: %s" (car files))))
         (commands (or (get backend 'more-vc-set-state-commands)
                       (error "No %s commands to set state" backend)))
         (ls (mapconcat (lambda (name)
                          (shell-quote-argument name))
                        files " "))
         (bufs (delq nil (mapcar 'find-buffer-visiting files))))
    ;; Each element in COMMANDS has the form (SHELL-COMMAND ARGS...).
    ;; SHELL-COMMAND is a string, and ARGS... are zero or more symbols, one
    ;; of `ls', which stands for the list FILES, shell-quoted and space-
    ;; separated, as a single string; or `state', which stands for STATE.
    (with-temp-buffer
      (dolist (command commands)
        (shell-command (eval (cons 'format command)) t)))
    (dolist (buf bufs)
      (with-current-buffer buf
        (revert-buffer t t)))
    bufs))

(put 'RCS 'more-vc-set-state-commands
     '(("co -l %s" ls)
       ("ci -u -s%S -m'State now %s.' %s" state state ls)))

(put 'CVS 'more-vc-set-state-commands
     '(("cvs commit -f -m'State now %s.' %s" state ls)
       ("cvs admin -s%S %s" state ls)
       ("rm -f %s" ls)
       ("cvs update %s" ls)))

;;;###autoload
(defun more-vc-make-sure-writeable ()
  "Make sure current buffer is writeable.
Do `vc-toggle-read-only' (perhaps twice!) if not.
Make sure underlying file is writeable, as well.
But all of this is done iff `this-command' is either
`add-change-log-entry' or `add-change-log-entry-other-window'."
  (when (memq this-command '(add-change-log-entry
                             add-change-log-entry-other-window))
    (flet ((try () (when buffer-read-only
                     (ttn-vc-toggle-read-only))))
      ;; twice in case vc-update intervened
      (try) (try))
    ;; vc was horribly crippled ca emacs 23
    (when buffer-read-only
      (toggle-read-only))
    (when (and (not buffer-read-only) buffer-file-name)
      (let ((mode (file-modes buffer-file-name)))
        (when (and mode (zerop (logand #o200 mode)))
          (set-file-modes buffer-file-name (logior #o200 mode)))))))

(defun more-vc-backend-get (property &optional filename)
  (let ((backend (vc-backend (or filename
                                 buffer-file-name
                                 (buffer-file-name vc-parent-buffer)))))
    (when backend
      (get backend property))))

;;;###autoload
(defun more-vc-pretty-up-log ()
  (buffer-disable-undo)
  (vc-exec-after (more-vc-backend-get 'more-vc-pretty-up-log-code)))

(put 'RCS 'more-vc-pretty-up-log-code
     '(let ((re (concat "\\(revision \\)\\S-+"
                        "\\(.*\ndate: \\)[^;]+"
                        "\\(;.+author: \\)[^;]+"
                        "\\(;.+state: \\)[^;]+"
                        "\\(\\(;.+lines: \\)[^;\n]+\\)*"
                        "\\(;.*\\)"))
            buffer-read-only)
        (flet ((mprop (n p v) (put-text-property (match-beginning n)
                                                 (match-end n)
                                                 p v)))

          (save-excursion
            (goto-char (point-min))
            (while (re-search-forward "^-+\n" nil t)
              (mprop 0 'display "\n\n")
              (looking-at re)
              (dolist (ent '((0 font-lock-face match)
                             (1 nil "  ")
                             (2) (3) (4) (6)
                             (7 nil "\n")))
                (let ((n (pop ent))
                      (p (or (pop ent) 'display))
                      (v (or (pop ent) "      ")))
                  (when (match-end n)
                    (mprop n p v)))))))
        (when (get-buffer-window (current-buffer))
          (recenter 0))))

(put 'CVS 'more-vc-pretty-up-log-code
     (get 'RCS 'more-vc-pretty-up-log-code))

(put 'Git 'more-vc-pretty-up-log-code
     '(let ((re (concat "\\(commit .*\n\\(Merge:.*\n\\)*Author:.*<\\)[^@]+"
                        "\\(@.*\nDate: +\\).*\n"))
            buffer-read-only)
        (flet ((mprop (n p v) (put-text-property (match-beginning n)
                                                 (match-end n)
                                                 p v)))

          (save-excursion
            (goto-char (point-min))
            (while (re-search-forward re nil t)
              (mprop 0 'font-lock-face 'match)
              (mprop 1 'display "  ")
              (mprop 3 'display "             "))))
        (when (get-buffer-window (current-buffer))
          (recenter 0))))

;;;###autoload
(defun more-vc-log-view-copy-id-as-kill ()
  "Copy the current log entry's \"id\" onto the `kill-ring'.
If point is not `looking-at' the `log-view-message-re', do nothing.
Otherwise, take the match's first group to be the entry's id, copy
this string onto the `kill-ring' (with `kill-new') and lastly,
display it in the echo area."
  (interactive)
  (when (looking-at log-view-message-re)
    (kill-new (buffer-substring-no-properties
               (match-beginning 1)
               (match-end 1)))
    (message "%s" (car kill-ring))))

;;;###autoload
(defun more-vc-likewise ()
  "Edit the current like to be \"\\t(TEXT): Likewise.\".
Leave point at eol.  When called, the line may have one of the forms:

  TEXT
  \\tTEXT
  (TEXT)
  \\t(TEXT)
  \\t* FILENAME (TEXT)

Here, \"\\t\" stands for the TAB character.  As a special case,
if unparenthesized TEXT (i.e., the first two cases) ends with
comma and any whitespace, that part is deleted.

This command is intended for use in Change Log mode."
  (interactive)
  (beginning-of-line)
  (delete-horizontal-space)
  (insert "\t")
  (unless (memq (char-after) '(?\( ?\*))
    (insert "(")
    (end-of-line)
    (delete-horizontal-space)
    (insert ")"))
  (end-of-line)
  (delete-horizontal-space)
  (when (= ?, (char-before))
    (delete-char -1) (insert ")"))
  (insert ": Likewise."))

;;;###autoload
(defun more-vc-diff-NEW ()
  "Diff current buffer's file with its \".NEW\" file.
For instance, if the current buffer is visiting \"foo\",
this command diffs \"foo\" and \"foo.NEW\"."
  (interactive)
  (let ((same-window-buffer-names (cons "*Diff*" same-window-buffer-names))
        (diff-default-read-only t)
        (short (file-name-nondirectory (buffer-file-name))))
    (diff short (concat short ".NEW") nil t)))

;;;###autoload
(defun more-vc-toggle-git-post-commit-hook ()
  "In .../.git/hooks/, toggle filename of post-commit hook.
Set to \"post-commit\" for enabled, \"post-commit.OFF\" for disabled."
  (interactive)
  (let (dir fn off)
    (if (not (setq dir (vc-git-root default-directory)))
        (message "Not under Git: %s" (directory-file-name default-directory))
      (setq fn (expand-file-name ".git/hooks/post-commit" dir)
            off (concat fn ".OFF"))
      (cond ((file-exists-p fn)
             (rename-file fn off)
             (message "%s DISABLED" fn))
            (t
             (rename-file off fn)
             (message "%s ENABLED" fn))))))

(defun more-vc-ref-at-point ()
  "Return the reference at point as a string.
This is essentially the symbol at point under Emacs Lisp syntax.
As a special case, if after the symbol there is \"@{...}\", (i.e.,
at-sign, open-curly-brace, ..., close-curly-brace), then that is
also included in the returned value.  See gitrevisions(7)."
  (save-excursion
    (with-syntax-table scheme-mode-syntax-table
      (forward-symbol 1)
      (let ((end (point)))
        (when (looking-at "@{[^{}]+}")
          (setq end (match-end 0)))
        (forward-symbol -1)
        (buffer-substring-no-properties (point) end)))))

;;;###autoload
(defun more-vc-copy-ref-as-kill ()
  "Copy the ref (see `more-vc-ref-at-point') to the kill ring."
  (interactive)
  (kill-new (more-vc-ref-at-point)))

;;;###autoload
(defun more-vc-git-rebase ()
  "Arrange to invoke `compile' interactively with a \"git rebase\" command.
The full command uses the REF (from `more-vc-ref-at-point') like so:
  git rebase -i REF^
Note the ^ (hat).  You can modify the command further prior to dispatch."
  (interactive)
  (let ((compile-command (format "git rebase -i --autosquash %s^"
                                 (more-vc-ref-at-point))))
    (call-interactively 'compile)))

;;;###autoload
(defun more-vc-git-show (&optional explicit-ref)
  "Do \"git show --stat -p REF\" (from `more-vc-ref-at-point').
Display output in a read-only buffer in Diff mode,
with Diff Auto-Refine mode enabled."
  (interactive "P")
  (let* ((ref (if explicit-ref
                  (read-string "Ref: ")
                (more-vc-ref-at-point)))
         (cmd (format "git show --stat -p %s" ref))
         (buf (get-buffer-create (format "*(%s) %s*"
                                         (file-name-nondirectory
                                          (directory-file-name
                                           (vc-git-root default-directory)))
                                         ref)))
         (resize-mini-windows nil)
         (pop-up-windows nil))
    (shell-command cmd buf)
    (with-current-buffer buf
      (diff-mode)
      (ignore-errors (diff-auto-refine-mode 1))
      (setq buffer-read-only t))
    (switch-to-buffer buf)))

(provide 'more-vc)

;;; more-vc.el ends here
